!=====================================================!
!                                                     !
!     This work is distributed under CC_BY_NC_SA      !
!                                                     !
! Created by Ruben M. Cabezon and Domingo Garcia-Senz !
!               ruben.cabezon@unibas.ch               !
!               domingo.garcia@upc.edu                !
!                                                     !
!                        (2018)                       !
!                                                     !
!=====================================================!
!                                                     !
!                  SPHYNX: update.f90                 !
!                                                     !
! Updates position, velocity and internal energy.     !
!=====================================================!

!----------------------------------------------------------------------------!
! This subroutine updates the position, velocities, internal energy and h of !
! all the particles. Uses a 2nd order centered scheme for integrating the    !
! equations.                                                                 !
!                                                                            !
! Velocities are updated in two ways depending on the variable 'relax'.      !
! If relax=.true. we are performing a relaxation procedure to let the        !
! particles to find their equilibrium position, erasing artificial pressure  !
! gradients derived from the initial random distribution of the particles.   !
! Thus particles are allowed to move over a sphere with constant radius      !
! and damping in velocity is applied to avoid unphysically large velocities. !
! In this case is better to calculate without gravity.                       !
!                                                                            !
! If relax=.false. we are in a normal SPH simulation.                        !
!............................................................................!

    SUBROUTINE update

      USE parameters

      IMPLICIT NONE
      DOUBLE PRECISION ax,ay,az
      DOUBLE PRECISION ux,uy,uz,vrad,vradx,vrady,vradz
      DOUBLE PRECISION aux,aux2,ax0,ay0,az0,val1,val2,val3
      DOUBLE PRECISION,DIMENSION(dim)::vcm

      INTEGER,DIMENSION(n-1)::indx3
      DOUBLE PRECISION b,yei,ti
      DOUBLE PRECISION,DIMENSION(n-1)::rd,yed,tempd,rdaux,yedaux,tempdaux
      DOUBLE PRECISION,DIMENSION(n)::radiusold

      INTEGER count,i,j,k,idmy
      LOGICAL endhere,corrected


      call profile(0)
      if(flags) write(*,*) 'Update'

!************************************************************************
!.. Standard SPH update.
!************************************************************************

      do j=0,dim-1
         do i=1,n
            ac(i+j*n)=-(gradp(i+j*n)-g*f(i+j*n))
         enddo
      enddo
         
!.. No movement for the first DELAY iterations to allow h stabilization.
      if(l.lt.iterini+delay) ac=0.d0   
      
      acmod=0.d0
      do j=0,dim-1
         do i=1,n
            acmod(i)=acmod(i)+ac(i+j*n)*ac(i+j*n)    !accel modulus(squared)
         enddo
      enddo
      do i=1,n
         acmod(i)=sqrt(acmod(i))
      enddo
         
         
      if(l.eq.iterini) then
         dtold=dt
         energyold(:)=energy(:)
         aviscold(:)=avisc(:)
         do j=0,dim-1
            do i=1,n
               a0(i+j*n)=a(i+j*n)-v(i+j*n)*dt
            enddo
         enddo
      endif
      
      aux=dt+dtold*.5d0
      aux2=(dt+dtold)*.5d0
      
      v0=v !save v_i to v_(i-1) before updating to v_(i+1)

!$omp parallel private(i,ax,ay,az,ax0,ay0,az0,val1,val2,val3)
!$omp do schedule(static)      
      do i=1,n
         ax=a(i)-despl                  !a_i
         ay=a(i+n)-despl
         az=a(i+n2)-despl
         
         ax0=a0(i)-despl                !a_(i-1)
         ay0=a0(i+n)-despl
         az0=a0(i+n2)-despl
         
         val1=(ax-ax0)/dtold
         val2=(ay-ay0)/dtold
         val3=(az-az0)/dtold
         
         v(i)=val1+ac(i)*aux            !v_(i+1)
         v(i+n)=val2+ac(i+n)*aux
         v(i+n2)=val3+ac(i+n2)*aux
         
         a0(i)=a(i)  !save a_i to a_(i-1) before updating to a_(i+1)
         a0(i+n)=a(i+n)
         a0(i+n2)=a(i+n2)
         
         a(i)=ax+val1*dt+(v(i)-val1)*dt*aux2/aux
         a(i+n)=ay+val2*dt+(v(i+n)-val2)*dt*aux2/aux
         a(i+n2)=az+val3*dt+(v(i+n2)-val3)*dt*aux2/aux
      enddo
!$omp end do

         

      if(ncubes.gt.1.and.l.ne.iterini) then
         !Assumes that everything is NOT CM-centered for PBC!!!

!$omp do schedule(static)
         do i=1,n                 !Check limits and applies PBC
            if(a(i).gt.xbox) then
               a(i)=a(i)-xbox
               a0(i)=a0(i)-xbox
            endif
            if(a(i+n).gt.ybox) then
               a(i+n)=a(i+n)-ybox
               a0(i+n)=a0(i+n)-ybox
            endif
            if(a(i+n2).gt.zbox) then
               a(i+n2)=a(i+n2)-zbox
               a0(i+n2)=a0(i+n2)-zbox
            endif
            if(a(i).lt.0.d0)then
               a(i)=a(i)+xbox
               a0(i)=a0(i)+xbox
            endif
            if(a(i+n).lt.0.d0)then
               a(i+n)=a(i+n)+ybox
               a0(i+n)=a0(i+n)+ybox
            endif
            if(a(i+n2).lt.0.d0)then
               a(i+n2)=a(i+n2)+zbox
               a0(i+n2)=a0(i+n2)+zbox
            endif
         enddo
!$omp end do
      endif

!$omp end parallel

      
!      call masscenter(n,1,n)
            
!            do j=0,dim-1
!               do i=1,n
!                  a(i+j*n)=a(i+j*n)-cm(j+1)
!               enddo
!            enddo
               

      val1=dt*dt*.5d0/dtold
      val2=dt+val1
      if(l.lt.iterini+delay) then
         val1=0.d0
         val2=0.d0
      endif

!$omp parallel private(i)
!$omp do schedule(static)      
      do i=1,n
         u(i)=u(i)+(.5d0*(energy(i)+avisc(i)))*val2-&  !U_(i+1)
              &(.5d0*(energyold(i)+aviscold(i)))*val1
      enddo
!$omp end do
!$omp end parallel


      do i=npini,npend
         if(u(i).le.0.d0.or.isnan(u(i))) then
            write(68,formatout) i,dble(l),a(i),a(i+n),h(i),dble(nvi(i)),temp(i),energy(i),avisc(i),energyold(i),aviscold(i),val1,val2,dt
         endif
      enddo
           


      if(l.ge.iterini+delay) then
            !For the next timestep:           !a_(i-1) is already done before.
         energyold(:)=energy(:)         !du/dt_(i-1)
         aviscold(:)=avisc(:)           !d(AV)/dt_(i-1)
         dtold=dt                       !dt_(i-1)
      endif

!      call masscenter(n,1,n)
      
      radius(:)=0.d0
!$omp parallel private(i)
!$omp do schedule(static)      
      do i=1,n
         radius(i)=sqrt(a(i)*a(i)+a(i+n)*a(i+n)+a(i+n2)*a(i+n2))
      enddo
!$omp end do
!$omp do schedule(static)   
      do i=1,n*dim
         a(i)=a(i)+despl
      enddo
!$omp end do
!$omp end parallel
      
      if(estabil)call estabilmod()!Creates radius(i) sorted with central dens

      !Smoothing length

!      if(l.gt.liniNR) then

!$omp parallel private(i)
      if(l.gt.(liniNR-5)) then
         if(volstdprom)then
!$omp do schedule(static)
            do i=1,n
               xmass(i)=(volstd(i))**pexp
            enddo
!$omp end do
         else
!$omp do schedule(static)
            do i=1,n
               xmass(i)=(masa(i)/promro(i))**pexp
            enddo
!$omp end do
         endif
      else
!$omp do schedule(static)
         do i=1,n
            xmass(i)=masa(i)
         enddo
!$omp end do
      endif
!$omp end parallel

      endhere=.false.
      do i=npini,npend
         if(isnan(a(i)).or.isnan(a(i+n))) then
            endhere=.true.
            write(*,*) i,'ISNAN!',l
            write(42,*) i,'ISNAN!',l
            write(42,formatin) a(i),a(i+n),a(i+n2),v(i),v(i+n),v(i+n2),&
                 &f(i),f(i+n),f(i+n2),gradp(i),gradp(i+n),gradp(i+n2),&
                 &v0(i),v0(i+n),v0(i+n2),ac(i),ac(i+n),ac(i+n2),&
                 &a0(i),a0(i+n),a0(i+n2),(cm(j),j=1,dim),&
                 &radius(i),masa(i),h(i),dble(nvi(i)),p(i),promro(i),u(i),ye(i),c(i)
         endif
      enddo
      if(endhere)stop

 12   format(1x,i6,30(1x,es12.5))
      call profile(1)

      RETURN
    END SUBROUTINE update
